/**
 * @file
 * VuoWindowOpenGLInternal implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#include "VuoWindowOpenGLInternal.hh"

#include <QCoreApplication>
#include <QGuiApplication>
#include <QResizeEvent>
#include <QScreen>
#include <QMenuBar>

#include <OpenGL/OpenGL.h>

extern "C"
{
#include "module.h"

#ifdef VUO_COMPILER
VuoModuleDetails({
					 "name" : "VuoWindowOpenGLInternal",
					 "dependencies" : [
						 "moc_VuoWindowOpenGLInternal.bc",
						 "VuoGLContext",
						 "ApplicationServices.framework",
						 "QtCore.framework",
						 "QtGui.framework",
						 "QtOpenGL.framework",
						 "QtWidgets.framework"
					 ]
				 });
#endif
}

/**
 * Creates a window containing an OpenGL view.
 *
 * Must be called on the main thread.
 */
VuoWindowOpenGLInternal::VuoWindowOpenGLInternal
(
		void (*initCallback)(void *),
		void (*resizeCallback)(void *, unsigned int width, unsigned int height),
		void (*drawCallback)(void *),
		void *drawContext
)
	: QMainWindow()
{
	this->callbacksEnabled = false;
	this->initCallback = initCallback;
	this->resizeCallback = resizeCallback;
	this->drawCallback = drawCallback;
	this->drawContext = drawContext;

	this->frameRequested = NULL;
	this->mouseMoved = NULL;
	this->mouseLeftPressed = NULL;

	this->aspect = -1;
	this->userResizedWindow = false;

	this->windowedGlContext = NULL;
	this->fullScreenGlContext = NULL;
	setCentralWidget(new QGLWidget());

	setWindowTitle("Vuo Scene");

	// Disable the close button.
	setWindowFlags((windowFlags() | Qt::CustomizeWindowHint) & ~Qt::WindowCloseButtonHint);

	isFullScreen = false;

	QMenu *viewMenu = new QMenu("View");
	fullScreenAction = viewMenu->addAction("Full Screen", this, SLOT(toggleFullScreen()), QKeySequence("Ctrl+F"));
	menuBar()->addMenu(viewMenu);

	exitFullScreenShortcut = new QShortcut(QKeySequence("Escape"), this, SLOT(toggleFullScreen()));
	exitFullScreenShortcut->setEnabled(false);

	this->resizingWindow = false;
	this->programmaticallyResizingWindow = true;
	resize(1024,768);
	this->programmaticallyResizingWindow = false;

	// We can't activate the OpenGL context until the first @c exposeEvent(),
	// so we initialize in @c render().
}

/**
 * Returns the QGLWidget inside this window.
 */
QGLWidget * VuoWindowOpenGLInternal::glWidget(void)
{
	return static_cast<QGLWidget *>(this->centralWidget());
}

/**
 * Returns the GL context belonging to the QGLWidget inside this window.
 */
QOpenGLContext * VuoWindowOpenGLInternal::glContext(void)
{
	return glWidget()->context()->contextHandle();
}

/**
 * Creates and initializes the GL Context (if it doesn't already exist), and activates it on the current thread.
 */
void VuoWindowOpenGLInternal::activateGLContext(void)
{
	bool shouldCreateContext;

	if (isFullScreen)
	{
		shouldCreateContext = (! fullScreenGlContext);
		if (shouldCreateContext)
		{
			createFullScreenGLContext();
			if (! fullScreenGlContext)
				fprintf(stderr, "VuoWindowOpenGLInternal::ensureGLContext() failed to create full-screen GL context.");

			// In case the window spans more than one display, hide it on the non-full-screen displays.
			// Use setWindowOpacity() instead of hide() in order to preserve the window's position.
			setWindowOpacity(0);
		}

		CGLSetCurrentContext((CGLContextObj) fullScreenGlContext);

		if (shouldCreateContext)
		{
			setFullScreenAspectRatio();
		}
	}
	else
	{
		if (fullScreenGlContext)
		{
			destroyFullScreenContext();
			setWindowOpacity(1);
		}

		shouldCreateContext = (! windowedGlContext);
		if (shouldCreateContext)
		{
			windowedGlContext = VuoGlContext_use();
			if (! glContext()->createWithNativeContext(windowedGlContext))
				fprintf(stderr, "VuoWindowOpenGLInternal::ensureGLContext() Failed to create windowed GL context.\n");
		}

		glWidget()->makeCurrent();
	}

	if (shouldCreateContext)
	{
		this->initCallback(this->drawContext);
	}
}

/**
 * Deactivates the GL context on the current thread.
 */
void VuoWindowOpenGLInternal::deactivateGLContext(void)
{
	if (isFullScreen)
	{
		glFlush();
		CGLFlushDrawable((CGLContextObj) fullScreenGlContext);
		CGLSetCurrentContext(NULL);
	}
	else
	{
		glWidget()->swapBuffers();
		glWidget()->doneCurrent();
	}
}

/**
 * Renders the OpenGL Window.
 */
void VuoWindowOpenGLInternal::render(void)
{
	if (! callbacksEnabled)
		return;

	activateGLContext();

	this->drawCallback(this->drawContext);

	deactivateGLContext();
}

/**
 * Override handling of certain events.
 */
bool VuoWindowOpenGLInternal::event(QEvent *event)
{
	switch (event->type())
	{
		case QEvent::UpdateRequest:
			render();
			return true;
		default:
			return QMainWindow::event(event);
	}
}

/**
 * Returns true if the window's actual aspect ratio differs by more than 1 pixel from the ideal aspect ratio.
 */
bool VuoWindowOpenGLInternal::actualAspectDiffersSignificantlyFromIdeal(void)
{
	int actualHeight = height();
	int idealHeight = width()/aspect;
	return abs(actualHeight - idealHeight) > 1;
}

/**
 * Render when the window is resized.
 */
void VuoWindowOpenGLInternal::resizeEvent(QResizeEvent *event)
{
	if (! callbacksEnabled || resizingWindow)
		return;

	// Avoid infinite recursion when this function calls resize() which calls this function.
	resizingWindow = true;

	if (aspect > 0 && actualAspectDiffersSignificantlyFromIdeal())
		resize(width(), width()/aspect);

	activateGLContext();

	// Set the viewport to window dimensions
	glViewport(0, 0, width()*devicePixelRatio(), height()*devicePixelRatio());

	// Let the caller recalculate the projection matrix
	this->resizeCallback(this->drawContext, width()*devicePixelRatio(), height()*devicePixelRatio());

	render();

	if (event && !this->programmaticallyResizingWindow)
		this->userResizedWindow = true;

	resizingWindow = false;
}

/**
 * Sets up the window to call trigger functions and other callbacks when events occur.
 */
void VuoWindowOpenGLInternal::enableTriggers
(
		VuoOutputTrigger(frameRequested, VuoInteger), /// @todo change to struct...
		VuoOutputTrigger(mouseMoved, VuoPoint2d),
		VuoOutputTrigger(mouseLeftPressed, VuoPoint2d)
)
{
	callbacksEnabled = true;
	resizeEvent(NULL);  // The initial resize event was ignored, so make one now. Without it, nothing gets rendered.

	this->frameRequested = (void (*)(VuoInteger)) frameRequested; /// @todo change to struct...
	this->mouseMoved = (void (*)(VuoPoint2d)) mouseMoved;
	this->mouseLeftPressed = (void (*)(VuoPoint2d)) mouseLeftPressed;
	/// @todo actually fire these triggers
}

/**
 * Stops the window from calling trigger functions and other callbacks when events occur.
 */
void VuoWindowOpenGLInternal::disableTriggers(void)
{
	callbacksEnabled = false;

	this->frameRequested = NULL;
	this->mouseMoved = NULL;
	this->mouseLeftPressed = NULL;
}

/**
 * Schedules the GL widget to be redrawn.
 */
void VuoWindowOpenGLInternal::scheduleRedraw(void)
{
	QCoreApplication::postEvent(this, new QEvent(QEvent::UpdateRequest));
}

/**
 * Sets the window's current and preferred aspect ratio.
 *
 * If necessary, the window is resized to match the specified aspect ratio.
 * After calling this method, when the user resizes the window, its height is adjusted to match the specified aspect ratio.
 */
void VuoWindowOpenGLInternal::setAspectRatio(unsigned int pixelsWide, unsigned int pixelsHigh)
{
	this->aspect = (float)pixelsWide/(float)pixelsHigh;
	if (actualAspectDiffersSignificantlyFromIdeal())
	{
		unsigned int desiredWidth;
		unsigned int desiredHeight;

		if (this->userResizedWindow)
		{
			// If the user manually resized the window, preserve its width.
			desiredWidth = width();
			desiredHeight = width()/aspect;
		}
		else
		{
			// Otherwise match the exact width and height provided by the calling function.
			desiredWidth = pixelsWide;
			desiredHeight = pixelsHigh;
		}

		// Set the desired size, unless it's larger than the window manager's allowed size.
		QSize availableSize = ((QGuiApplication *)QGuiApplication::instance())->primaryScreen()->availableSize();
		float margin = 0.99;
		this->programmaticallyResizingWindow = true;
		if (desiredWidth <= availableSize.width()*margin && desiredHeight <= availableSize.height()*margin)
			resize(desiredWidth, desiredHeight);
		else if (desiredWidth > availableSize.width()*margin)
			// Too wide, so scale it to almost the maximum horizontal screen space.
			resize(availableSize.width()*margin, availableSize.width()*margin/aspect);
		else
			// Too tall, so scale it to almost the maximum vertical screen space.
			resize(availableSize.height()*margin*aspect, availableSize.height()*margin);
		this->programmaticallyResizingWindow = false;
	}
}

/**
 * Toggles between full-screen and windowed display.
 */
void VuoWindowOpenGLInternal::toggleFullScreen()
{
	isFullScreen = ! isFullScreen;

	fullScreenAction->setText(isFullScreen ? "Windowed" : "Full Screen");
	exitFullScreenShortcut->setEnabled(isFullScreen);

	render();
}

/**
 * Creates a full-screen GL context for the monitor showing the majority of the window's surface area.
 */
void VuoWindowOpenGLInternal::createFullScreenGLContext(void)
{
	// Find the display that shows the majority of the window's surface area.

	const uint32_t kMaxDisplays = 1024;
	CGDirectDisplayID displays[kMaxDisplays];
	CGDisplayCount displayCount;
	CGGetActiveDisplayList(kMaxDisplays, displays, &displayCount);

	CGFloat maxIntersectionArea;
	CGRect windowRect = CGRectMake(x(), y(), width(), height());
	for (uint32_t i = 0; i < displayCount; ++i)
	{
		CGRect displayRect = CGDisplayBounds(displays[i]);
		CGRect intersectionRect = CGRectIntersection(windowRect, displayRect);
		CGFloat intersectionArea = (CGRectIsNull(intersectionRect) ?
										0 :
										CGRectGetWidth(intersectionRect) * CGRectGetHeight(intersectionRect));
		if (intersectionArea > maxIntersectionArea)
		{
			maxIntersectionArea = intersectionArea;
			fullScreenDisplay = displays[i];
		}
	}

	// Create the full-screen context and do other setup.
	// https://developer.apple.com/library/mac/samplecode/GLFullScreen

	CGDisplayCapture(fullScreenDisplay);

	CGLPixelFormatAttribute attribs[] =
	{
		// The full-screen attribute
		kCGLPFAFullScreen,

		// The display mask associated with the captured display
		// We may be on a multi-display system (and each screen may be driven by a different renderer), so we need to specify which screen we want to take over.
		kCGLPFADisplayMask, (CGLPixelFormatAttribute) CGDisplayIDToOpenGLDisplayMask(fullScreenDisplay),

		// Attributes common to full-screen and non-fullscreen
		kCGLPFAAccelerated,
		kCGLPFANoRecovery,
		kCGLPFADoubleBuffer,
		kCGLPFAColorSize, (CGLPixelFormatAttribute) 24,
		kCGLPFADepthSize, (CGLPixelFormatAttribute) 16,
		(CGLPixelFormatAttribute) 0
	};

	// By specifying the non-fullscreen context as the shareContext, we automatically inherit the OpenGL objects (textures, etc) it has defined.
	CGLPixelFormatObj pixelFormatObj;
	GLint numPixelFormats;
	CGLChoosePixelFormat(attribs, &pixelFormatObj, &numPixelFormats);
	CGLCreateContext(pixelFormatObj, (CGLContextObj) windowedGlContext, (CGLContextObj*) &fullScreenGlContext);
	CGLDestroyPixelFormat(pixelFormatObj);

	if (! fullScreenGlContext)
	{
		CGDisplayRelease(fullScreenDisplay);
		return;
	}

#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdeprecated-declarations"
	// Attach a full-screen drawable object to the current context
	CGLSetFullScreenOnDisplay((CGLContextObj) fullScreenGlContext, CGDisplayIDToOpenGLDisplayMask(fullScreenDisplay));
#pragma clang diagnostic pop
}

/**
 * Destroys the full-screen GL context.
 */
void VuoWindowOpenGLInternal::destroyFullScreenContext(void)
{
	CGLSetCurrentContext(NULL);
	CGLClearDrawable((CGLContextObj) fullScreenGlContext);
	CGLDestroyContext((CGLContextObj) fullScreenGlContext);
	CGDisplayRelease(fullScreenDisplay);

	fullScreenGlContext = NULL;
}

/**
 * Set the aspect ratio for the full-screen display to be the same as the window's.
 *
 * Assumes the full-screen GL context is the current context.
 */
void VuoWindowOpenGLInternal::setFullScreenAspectRatio(void)
{
	float windowAspectRatio = width() / (float) height();

	CGRect displayRect = CGDisplayBounds(fullScreenDisplay);
	float displayAspectRatio = CGRectGetWidth(displayRect) / CGRectGetHeight(displayRect);

	GLint viewportX, viewportY;
	GLsizei viewportWidth, viewportHeight;
	if (windowAspectRatio < displayAspectRatio)
	{
		// Window's aspect ratio is narrower than display's, so letterbox on left and right.
		viewportHeight = CGRectGetHeight(displayRect);
		viewportWidth = windowAspectRatio * viewportHeight;
		viewportX = (CGRectGetWidth(displayRect) - viewportWidth) / 2;
		viewportY = 0;
	}
	else
	{
		// Window's aspect ratio is wider than display's, so letterbox on top and bottom.
		viewportWidth = CGRectGetWidth(displayRect);
		viewportHeight = viewportWidth / windowAspectRatio;
		viewportY = (CGRectGetHeight(displayRect) - viewportHeight) / 2;
		viewportX = 0;
	}

	glViewport(viewportX*devicePixelRatio(), viewportY*devicePixelRatio(), viewportWidth*devicePixelRatio(), viewportHeight*devicePixelRatio());
}
