﻿/**
 * @file
 * Prototypes for node class implementations.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#ifndef NODE_H
#define NODE_H


#include "module.h"

#include <stdbool.h>
#include <stdio.h>
#include <unistd.h>

// Types
#include "VuoColor.h"
#include "VuoInteger.h"
#include "VuoImage.h"
#include "VuoMidiController.h"
#include "VuoMidiDevice.h"
#include "VuoMidiNote.h"
#include "VuoPoint2d.h"
#include "VuoPoint3d.h"
#include "VuoPoint4d.h"
#include "VuoReal.h"
#include "VuoSceneObject.h"
#include "VuoVertices.h"
#include "VuoShader.h"
#include "VuoText.h"
#include "VuoTransform.h"
#include "VuoGradientNoise.h"
#include "VuoBlendMode.h"
#include "VuoBoolean.h"
#include "VuoCountWrapMode.h"
#include "VuoCurveDomain.h"
#include "VuoNoise.h"
#include "VuoCurve.h"
#include "VuoMouseButtonAction.h"

// List Types
#include "VuoList_VuoBoolean.h"
#include "VuoList_VuoImage.h"
#include "VuoList_VuoInteger.h"
#include "VuoList_VuoMidiDevice.h"
#include "VuoList_VuoPoint2d.h"
#include "VuoList_VuoPoint3d.h"
#include "VuoList_VuoPoint4d.h"
#include "VuoList_VuoReal.h"
#include "VuoList_VuoSceneObject.h"
#include "VuoList_VuoVertices.h"
#include "VuoList_VuoText.h"


/**
 * @addtogroup WritingNodeClasses
 * @{
 */


/**
 * @defgroup VuoNodeInstanceData Node Instance Data
 * Node instance data structure.
 *
 * @{
 */

/**
 * @struct nodeInstanceData
 * Stateful nodes may define a custom structure called @c nodeInstanceData, to store the state of a single node instance.
 */
struct nodeInstanceData;

/**
 * @}
 */



/**
 * @defgroup VuoNodeParameters Node Parameters
 * Parameter decorations to be used by node classes.
 *
 * @{
 */

/**
 * Conductivity options for input ports.
 */
enum VuoPortConductivity
{
	/**
	 * An event received by this input port is conducted to all output ports.
	 */
	VuoPortAlwaysConducts,

	/**
	 * An event received by this input port may be conducted to all, some, or none of the output ports.
	 */
	VuoPortMayConduct,

	/**
	 * An event received by this input port is never conducted to any output port.
	 */
	VuoPortNeverConducts
};

/**
 * Use this to decorate parameters referring to a stateful node's instance data structure.
 * @hideinitializer
 *
 * @param type The instance data's C type.
 *
 * \eg{void nodeInstanceEvent(VuoInstanceData(struct nodeInstanceData) ctx);}
 */
#define VuoInstanceData(type) __attribute__((annotate("vuoInstanceData"))) type * const

/**
 * Use this to decorate parameters acting as data input ports.  The input port name should use @c lowerCamelCase.
 * @hideinitializer
 *
 * @param type This input port's C type. Must be a typedef from a Vuo type definition.
 * @param ... Optionally, a JSON object specification containing additional details about the input port's data, such as the default value.
 *
 * \eg{void nodeEvent(VuoInputData(VuoInteger,{"default":60,"suggestedMin":0,"suggestedMax":127}) noteNumber);}
 */
#define VuoInputData(type, ...) __attribute__((annotate("vuoInputData"),annotate("vuoType:" #type),annotate("vuoInputDataDetails: " #__VA_ARGS__))) const type

/**
 * Use this to decorate parameters acting as event input ports.  Indicates whether this particular input port was pushed.  The input port name should use @c lowerCamelCase.
 * @hideinitializer
 *
 * @param conductivity A value of type VuoPortConductivity, indicating the port's type of conductivity.
 * @param name The identifier of the corresponding data port, or blank if this is intended to be an event-only input port.
 *
 * \eg{void nodeEvent(VuoInputData(VuoInteger,"0") seconds, VuoInputConductor(VuoPortNeverConducts,seconds) secondsPushed);}
 * \eg{void nodeEvent(VuoInputConductor(VuoPortAlwaysConducts,) start);}
 */
#define VuoInputConductor(conductivity,name) __attribute__((annotate("vuoInputConductivity:" #conductivity),annotate("vuoInputConductor:" #name))) const bool

/**
 * Use this to decorate parameters acting as data output ports.  The output port name should use @c lowerCamelCase.
 * @hideinitializer
 *
 * @param type This output port's C type. Must be a typedef from a Vuo type definition.
 *
 * \eg{void nodeEvent(VuoOutputData(VuoInteger) seconds);}
 */
#define VuoOutputData(type) __attribute__((annotate("vuoOutputData"), annotate("vuoType:" # type))) type * const

/**
 * Use this to decorate parameters acting as event output ports.  Indicates whether a push should be conducted through this particular output port.  The output port name should use @c lowerCamelCase.
 *
 * @param name The identifier of the corresponding data port, or blank if this is intended to be an event-only output port.
 *
 * \eg{void nodeEvent(VuoOutputData(VuoInteger) seconds, VuoOutputConductor(seconds) secondsPush);}
 * \eg{void nodeEvent(VuoOutputConductor() started);}
 */
#define VuoOutputConductor(name) __attribute__((annotate("vuoOutputConductor:" #name))) bool * const

/**
 * Use this to decorate parameters acting as event triggers.  Indicates whether an event should be fired from this particular output port.
 * @hideinitializer
 *
 * @param name The identifier of the triggers function.  Use @c lowerCamelCase.
 * @param type The C type of the data passed along when an event is fired.
 *
 * \eg{void nodeEvent(VuoOutputTrigger(started,void))
 * {
 *     // Generate an event without any data.
 *     started();
 * }}
 * \eg{void nodeEvent(VuoOutputTrigger(ding,int))
 * {
 *     // Generate an event with stack data.
 *     ding(5);
 * }}
 * \eg{void nodeEvent(VuoOutputTrigger(ding,char *))
 * {
 *     // Generate an event with heap data.
 *     char *heapData = (char *)malloc(strlen("hello")+1);
 *     strcpy(heapData, "hello");
 *     ding(heapData);
 *     VuoRelease(heapData);
 * }}
 */
#define VuoOutputTrigger(name,type) __attribute__((annotate("vuoOutputTrigger:" #name), annotate("vuoType:" # type))) void (* const name)(type)

/**
 * @}
 */



// These prototypes should only appear to Doxygen, since individual node classes' parameter lists can vary.
#ifdef DOXYGEN

/**
 * @defgroup VuoNodeMethodsStateless Node Methods: Stateless
 * Event handler method to be implemented by node classes.
 *
 * @{
 */

/**
 * Called for every push this node class receives.
 *
 * Parameter decorations may include:
 *		@arg @ref VuoInputData
 *		@arg @ref VuoInputConductor
 *		@arg @ref VuoOutputTrigger
 *		@arg @ref VuoOutputData
 *		@arg @ref VuoOutputConductor
 */
void nodeEvent(...);

/**
 * @}
 */



/**
 * @defgroup VuoNodeMethodsStateful Node Methods: Stateful
 * Setup, event handler, and teardown methods to be implemented by stateful node classes.
 *
 * @{
 */

/**
 * Called when a new instance of the node class is created.
 * The implementation should allocate, initialize, register, and return whatever instance data is required.
 * It may open services (file, network port, ...).
 *
 * Parameter decorations may include:
 *		@arg @ref VuoInputData — This argument has the initial (constant or default) value of the input port, before it has received any events.
 */
struct nodeInstanceData *nodeInstanceInit(...);

/**
 * Called just after @c nodeInstanceInit() and each time the composition is updated during live coding.
 * The implementation may start any dispatch sources or event loops that use @c VuoOutputTrigger functions as callbacks.
 *
 * This function is optional.
 *
 * Parameter decorations should include at least:
 *		@arg @ref VuoInstanceData
 *
 * Parameter decorations may include:
 *		@arg @ref VuoInputData — This argument has the current value of the input port.
 *		@arg @ref VuoOutputTrigger — Stateful nodes may store references to trigger functions in their instance data, in order to generate events at any time.
 */
void nodeInstanceCallbackStart(...);

/**
 * Called each time this node instance receives an event.
 *
 * Parameter decorations should include at least:
 *		@arg @ref VuoInstanceData
 *
 * Parameter decorations may include:
 *		@arg @ref VuoInputData
 *		@arg @ref VuoInputConductor
 *		@arg @ref VuoOutputTrigger
 *		@arg @ref VuoOutputData
 *		@arg @ref VuoOutputConductor
 */
void nodeInstanceEvent(...);

/**
 * Called each time an input port constant on this node instance is updated during live coding.
 * The implementation should adjust all @c VuoOutputTrigger callbacks to use the new input port values.
 *
 * This function is optional.
 *
 * Parameter decorations should include at least:
 *		@arg @ref VuoInstanceData
 *
 * Parameter decorations may include:
 *		@arg @ref VuoInputData — This argument has the current value of the input port.
 *		@arg @ref VuoOutputTrigger — Stateful nodes may store references to trigger functions in their instance data, in order to generate events at any time.
 */
void nodeInstanceCallbackUpdate(...);

/**
 * Called each time the composition is updated during live coding and just before @c nodeInstanceFini().
 * The implementation should stop all @c VuoOutputTrigger callbacks started in @c nodeInstanceCallbackStart() or elsewhere.
 * Any @c VuoOutputTrigger functions stored in instance data become invalid after this function returns.
 *
 * This function is optional.
 */
void nodeInstanceCallbackStop(struct nodeInstanceData * const context);

/**
 * Called when the specified node instance is no longer needed.
 * The implementation should close any services opened in @c nodeInstanceInit() or elsewhere.
 */
void nodeInstanceFini(struct nodeInstanceData * const context);

/**
 * @}
 */

#endif


/**
 * @}
 */


#endif
